# Copyright Louis Paternault 2019-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Read and write PDF files using the PyMuPDF library."""

import contextlib
import decimal
import io

import pymupdf

# pylint: disable=wrong-import-position
from . import AbstractPDFFileReader, AbstractPDFFileWriter, AbstractPDFPage

VERSION = "pymupdf" + " ".join(pymupdf.version[0])


class PDFFileReader(AbstractPDFFileReader):
    """Read a PDF file."""

    def __init__(self, file=None):
        super().__init__()
        if isinstance(file, io.BytesIO):
            self._file = pymupdf.open(stream=file, filetype="application/pdf")
        else:
            self._file = pymupdf.open(file)

    def close(self):
        self._file.close()

    def __len__(self):
        return self._file.page_count

    def __iter__(self):
        for page in self._file:
            yield PDFPage(page)

    def __getitem__(self, key):
        return self._file[key]

    @property
    def metadata(self):
        # pylint: disable=no-member
        return self._file.metadata


class PDFFileWriter(AbstractPDFFileWriter):
    """PDF file writer."""

    def __init__(self):
        super().__init__()
        self._file = pymupdf.Document()

    @contextlib.contextmanager
    def new_page(self, width, height):
        yield PDFPage(self._file.new_page(width=width, height=height))

    def write(self, file):
        if isinstance(file, io.BytesIO):
            file.write(self._file.write())
        else:
            self._file.save(file)

    @property
    def metadata(self):
        # pylint: disable=no-member
        return self._file.metadata

    @metadata.setter
    def metadata(self, value):
        self._file.set_metadata(value)


class PDFPage(AbstractPDFPage):
    """Page of a PDF file (using PyMuPDF)."""

    @property
    def parent(self):
        """PDF Object this page belongs to."""
        return self._page.parent

    @property
    def number(self):
        """Number of this page in the PDF."""
        return self._page.number

    @property
    def mediabox_size(self):
        """Return size of the mediabox, as a tuple of `(width, height)`."""
        return self._page.mediabox_size

    @property
    def rotated_size(self):
        if self._page.rotation % 180 == 0:
            return self._page.mediabox_size
        return (
            self._page.mediabox_size[1],
            self._page.mediabox_size[0],
        )

    def merge_translated_page(self, page, x, y):
        if (
            x + decimal.Decimal(page.mediabox_size.x) < 0
            or x > self.mediabox_size.x
            or y + decimal.Decimal(page.mediabox_size.y) < 0
            or y > self.mediabox_size.y
        ):
            # The merged page is drawn outside the current page (empty intersection).
            # Method show_pdf_page() does not like it, and raise an exception.
            return

        rotation = page.rotation
        page.set_rotation(0)
        if rotation % 180 == 0:
            width = page.mediabox_size[0]
            height = page.mediabox_size[1]
        else:
            width = page.mediabox_size[1]
            height = page.mediabox_size[0]

        self._page.show_pdf_page(
            pymupdf.Rect(
                x,
                decimal.Decimal(self.mediabox_size[1]) - y - decimal.Decimal(height),
                x + decimal.Decimal(width),
                decimal.Decimal(self.mediabox_size[1]) - y,
            ),
            page.parent,
            page.number,
            rotate=rotation,
        )



# Copyright 2014-2025 Louis Paternault and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Paper-size related functions"""

import os
import subprocess
import typing
from decimal import Decimal

import papersize

from . import errors


def parse_lc_paper(string) -> tuple[Decimal, Decimal]:
    """Parse LC_PAPER locale variable

    We assume units are milimeters.
    """
    dimensions = {}
    for line in string.split("\n"):
        if line.startswith("width="):
            dimensions["width"] = papersize.parse_length(f"{line[6:]}mm")
        if line.startswith("height="):
            dimensions["height"] = papersize.parse_length(f"{line[7:]}mm")
    if len(dimensions) == 2:
        return (dimensions["width"], dimensions["height"])
    raise errors.CouldNotParse(string)


def target_papersize(target_size) -> tuple[Decimal, Decimal]:
    """Return the target paper size.

    :param str|tuple[decimal.Decimal]|None target_size: Target size,
        if provided by user in command line.
    """
    # pylint: disable=too-many-return-statements

    # Option set by user on command line
    if target_size is not None:
        if isinstance(target_size, str):
            return papersize.parse_papersize(target_size)
        return target_size

    # LC_PAPER environment variable (can be read from "locale -k LC_PAPER"
    try:
        return parse_lc_paper(
            subprocess.check_output(["locale", "-k", "LC_PAPER"], text=True)
        )
    except (FileNotFoundError, subprocess.CalledProcessError, errors.CouldNotParse):
        pass

    # PAPERSIZE environment variable
    try:
        return papersize.parse_papersize(os.environ["PAPERSIZE"].strip())
    except KeyError:
        pass

    # file described by the PAPERCONF environment variable
    try:
        return papersize.parse_papersize(
            # pylint: disable=unspecified-encoding
            open(os.environ["PAPERCONF"])
            .read()
            .strip()
        )
    except errors.CouldNotParse:
        raise
    except:  # pylint: disable=bare-except
        pass

    # content of /etc/papersize
    try:
        # pylint: disable=unspecified-encoding
        return papersize.parse_papersize(open("/etc/papersize").read().strip())
    except errors.CouldNotParse:
        raise
    except:  # pylint: disable=bare-except
        pass

    # stdout of the paperconf command
    try:
        return papersize.parse_papersize(
            subprocess.check_output(["paperconf"], text=True).strip()
        )
    except (FileNotFoundError, subprocess.CalledProcessError, errors.CouldNotParse):
        pass

    # Eventually, if everything else has failed, a4
    return papersize.parse_papersize("a4")


class Margins(typing.NamedTuple):
    """Represents the margins at all four sides of a page."""

    top: Decimal
    right: Decimal
    bottom: Decimal
    left: Decimal

    @classmethod
    def parse(
        cls,
        margin: "None | str | Decimal | typing.Sequence[str | Decimal] | typing.Self",
    ) -> "typing.Self":
        """Normalize a margin specification into a (top, right, bottom, left) quadruple
        and parse any strings into numbers.
        """
        match margin:
            case Margins(top, right, bottom, left):
                pass
            case None:
                return cls.parse(papersize.parse_length(string="0"))
            case str() as margin:
                return cls.parse((papersize.parse_length(margin),))
            case Decimal() as margin:
                return cls.parse((margin,))
            case [*margins]:  # "sequence" pattern despite square-brackets
                match tuple(
                    papersize.parse_length(m) if not isinstance(m, Decimal) else m
                    for m in margins
                ):
                    case (all_four_sides,):
                        top = right = bottom = left = all_four_sides
                    case (
                        vertical,
                        horizontal,
                    ):
                        top = bottom = vertical
                        right = left = horizontal
                    case (
                        top,
                        horizontal,
                        bottom,
                    ):
                        left = right = horizontal
                    case (
                        top,
                        right,
                        bottom,
                        left,
                    ):
                        pass
                    case _:
                        return cls.parse(papersize.parse_length(string="0"))
        return cls(top, right, bottom, left)



# Copyright 2014-2025 Louis Paternault and contributors

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Convert PDF files to 'n-up' PDF files, guessing the output layout."""

import collections
import decimal
import io
import logging
import math
import pathlib
from collections.abc import Callable, Sequence
from typing import Any, Literal

from . import errors, geometry, paper, pdfbackend

VERSION = "1.13.0"
__AUTHOR__ = "Louis Paternault (spalax@gresille.org)"
__COPYRIGHT__ = "(C) 2014-2025 Louis Paternault. GNU AGPL 3 or later."

LOGGER = logging.getLogger("pdfautonup")


def _none_function(*args, **kwargs):  # pylint: disable=unused-argument
    """Accept any number of arguments. and does nothing."""


class _PageSequence(collections.abc.Sequence):
    """Sequence of pages of several PDF files."""

    def __init__(self, filenames):
        self.files = []
        self._filenames = filenames

    def __enter__(self):
        for name in self._filenames:
            try:
                if name == "-":
                    self.files.append(pdfbackend.get_backend().PDFFileReader())
                else:
                    self.files.append(pdfbackend.get_backend().PDFFileReader(name))
            except (FileNotFoundError, PermissionError) as error:
                raise errors.PdfautonupError(
                    f"Error while reading file '{name}': {error}."
                )
            except RuntimeError as error:
                raise errors.PdfautonupError(
                    f"Error: Malformed file '{name}': {error}."
                )
        return self

    def __exit__(self, *exc):
        for file in self.files:
            file.close()

    def __iter__(self):
        for pdf in self.files:
            yield from pdf

    def __len__(self):
        return sum(len(pdf) for pdf in self.files)

    def __getitem__(self, index):
        for file in self.files:
            try:
                return file[index]
            except IndexError:
                index -= len(file)
        raise IndexError

    def get_metadata(self):
        """Aggregate metadata from input files."""
        if len(self.files) == 1:
            return self.files[0].metadata

        input_info = [pdf.metadata for pdf in self.files]
        output_info = {}
        for key in pdfbackend.METADATA_KEYS:
            values = list(
                data[key]
                for data in input_info
                if (key in data and (data[key] is not None))
            )
            if values:
                output_info[key] = " / ".join([f"“{item}”" for item in values])
        return output_info


def pdfautonup(
    files: Sequence[str | pathlib.Path | io.BytesIO],
    output: str | pathlib.Path | io.BytesIO,
    size: str | None | tuple[decimal.Decimal] = None,
    *,
    algorithm: Literal["fuzzy", "panel"] | None = None,
    repeat: Literal["auto", "fit"] | int = "auto",
    more: dict[str, Any] = None,
    orientation: Literal["auto", "portrait", "landscape"] = "auto",
    progress: Callable[[int, int], None] = _none_function,
):  # pylint: disable=too-many-arguments, too-many-locals
    """Convert PDF files to 'n-up' PDF files, guessing the output layout.

    .. versionadded:: 1.9.0

    .. versionchanged:: 1.12.0
       Source and output files can be of type :class:`io.BytesIO`.

    .. versionchanged:: 1.12.0
       Removed `interactive` parameter.

    .. versionchanged:: 1.13.0
       Argument `more` now accepts a `"writingorder"` key
       (see command line option `--writing-order`).

    :param files: Files to process.
        It is the caller responsability to call :meth:`io.BytesIO.seek` before
        calling this function, if necessary.
    :param output: Output file.
        It is the caller responsability to call :meth:`io.BytesIO.seek` before
        calling this function, if necessary.
    :param size: Size of the pages of the destination file, either as:
        a tuple of :class:`decimal.Decimal` (width and height, in points);
        a string to be parsed by :func:`papersize.parse_papersize`;
        :class:`None`, in which case the default paper size will be used.
    :param repeat: If a number, repeat the input file this number of times.
        If ``"fit"``, repeat the input files as many times as necessary to fill
        all pages of the destination file. If ``"auto"``, is equivalent to
        ``"fit"`` if input file has one page, and equivalent to ``1``
        otherwise.
    :param algorithm: Select one algorithm, either ``"fuzzy"`` (document pages
        can overlap or leave blank space between them, but not too much) or
        ``"panel"`` (the gap length between source pages is fixed, and a
        minimum destination margin is respected). If ``None``, chooses
        ``"panel"`` if ``"margin"`` or ``"gap"`` are defined in ``more``, and
        ``"fuzzy"`` otherwise.
    :param orientation: Force orientation of destination file to portrait or landscape.
        If ``"auto"``, select the one that fits the most input pages.
    :param more: Additional arguments for algorithms.
        Both algorithms accept argument ``writingorder``, and the *panel*
        algorithm also accepts ``"margin"`` and ``"gap"`` (as strings
        to parsed by :func:`papersize.parse_length` or :class:`decimal.Decimal`
        as the length in points).
        ``"margin"`` may also be a 2- to 4-tuple to specify individual margins
        in the same order and arity as in CSS.
    :param progress: A function that takes to integer arguments
        (number of pages processed so far, and total number of pages to
        process), and display a progress. Whether it is displayed on standard
        input or a GUI or something else is up to you.

    .. warning::

       If a file is ``"-"``, it is read from standard input.
       I am still undecided about how reading from standard input should be
       handled, so this might change in the future.

       So, right now, reading from standard input is unsupported, and input
       files cannot be named ``"-"``.
    """
    # pylint: disable=too-many-branches

    with _PageSequence(files) as pages:
        if not pages:
            raise errors.PdfautonupError("Error: PDF files have no pages to process.")

        page_sizes = list(zip(*[page.rotated_size for page in pages]))
        source_size = (
            decimal.Decimal(max(page_sizes[0])),
            decimal.Decimal(max(page_sizes[1])),
        )
        target_size = paper.target_papersize(size)

        if [len(set(page_sizes[i])) for i in (0, 1)] != [1, 1]:
            LOGGER.warning(
                "Pages have different sizes. The result might be unexpected."
            )

        if more is None:
            more = {}
        if algorithm is None:
            if more.get("gap", None) is None and more.get("margin", None) is None:
                fit = geometry.Fuzzy
            else:
                fit = geometry.Panelize
        else:
            fit = {"fuzzy": geometry.Fuzzy, "panel": geometry.Panelize}[algorithm]

        dest = fit(
            source_size,
            target_size,
            orientation=orientation,
            more=more,
        )

        if repeat == "auto":
            if len(pages) == 1:
                repeat = "fit"
            else:
                repeat = 1
        if repeat == "fit":
            repeat = math.lcm(dest.pages_per_page, len(pages)) // len(pages)

        totalpages = repeat * len(pages)
        progress(0, totalpages)
        for destcount in range(math.ceil(totalpages / dest.pages_per_page)):
            with dest.new_page() as destpage:
                for sourcecount in range(
                    destcount * dest.pages_per_page,
                    (destcount + 1) * dest.pages_per_page,
                ):
                    if sourcecount < totalpages:
                        dest.add_page(
                            pages[sourcecount % len(pages)],
                            destpage,
                            sourcecount % dest.pages_per_page,
                        )
                        progress(sourcecount + 1, totalpages)

        dest.set_metadata(pages.get_metadata())
        dest.write(output)



# Copyright 2014-2024 Louis Paternault and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Manage options"""

import argparse
import os
import sys
import textwrap

import papersize

from . import VERSION, order, pdfbackend
from .paper import Margins


def length_type(text):
    """Check type of length (number plus optional unit).

    Wrapper to :func:`papersize.parse_length`.
    """
    try:
        return papersize.parse_length(text)
    except papersize.CouldNotParse as error:
        raise argparse.ArgumentTypeError(str(error))


def margin_type(text: str) -> Margins:
    """Check type of margin specifier (one to four lengths separated by spaces)."""
    margins = text.split()
    if not 1 <= len(margins) <= 4:
        raise argparse.ArgumentTypeError(
            f"expected 1 to 4 space-separated lengths for margin, found {len(margins)}"
        )
    try:
        return Margins.parse(margins)
    except papersize.CouldNotParse as error:
        raise argparse.ArgumentTypeError(str(error))


def size_type(text):
    """Check type of paper size (couple of numbers plus optional units).

    Wrapper to :func:`papersize.parse_papersize`.
    """
    try:
        return papersize.parse_papersize(text)
    except papersize.CouldNotParse as error:
        raise argparse.ArgumentTypeError(str(error))


def repeat_type(text):
    """Check type of '--repeat' option.

    Must be either a positive integer, or 'fit' or 'auto'.
    """
    if text in ["auto", "fit"]:
        return text
    try:
        if int(text) > 0:
            return int(text)
        raise ValueError
    except ValueError as error:
        raise argparse.ArgumentTypeError(
            textwrap.dedent(
                """
        Argument must be either 'fit' or 'auto', or a positive integer.
        """
            )
        ) from error


def progress_type(text):
    """Return plain progress text, turning aliases into their value."""
    return {
        "dot": ".",
        "percent": "{percent}%\n",
        "pages": "{page}/{total}\n",
        "none": "",
    }.get(text, text)


def inputfile_type(filename):
    """Return the argument, with the `.pdf` extension if missing (and if argument is not "-").

    If `filename` does not exist, but `filename.pdf` does exist, return the
    latter. Otherwise (even if it does not exist), return the former.
    """
    if filename == "-":
        return None
    if not os.path.exists(filename):
        extended = f"{filename}.pdf"
        if os.path.exists(extended):
            return extended
    return filename


class ArgparseAction(argparse.Action):
    """Argparse action to display help."""

    #: pylint: disable=too-few-public-methods, abstract-method

    def __init__(self, *args, **kwargs):
        if "nargs" in kwargs:
            raise ValueError("nargs not allowed")
        kwargs["nargs"] = 0
        super().__init__(*args, **kwargs)


class HelpPaper(ArgparseAction):
    """Argparse action to display help about paper sizes."""

    #: pylint: disable= line-too-long

    def __call__(self, *args, **kwargs):
        # pylint: disable=signature-differs
        print(
            textwrap.dedent(
                """
            # Source

            Paper size is read from the following sources (in that order):

            - Argument of "--size" option;
            - LC_PAPER environment variable (read as mm);
            - PAPERSIZE environment variable;
            - content of file specified by the PAPERCONF environment variable;
            - content of file /etc/papersize;
            - output of the paperconf command;
            - if everything else have failed, A4.

            # Units

            Argument of --margin and --gap is a number and a unit (e.g. "1.2mm"). If no unit is given, the default unit is the point (1/72 of an inch, or about 0.353mm).

            Argument of --margin may also be up to 4 numbers to specify margin on each side, in the same way as CSS. Some extra margin may remain on the bottom.

            Available units are: {units}. See `pdfautonup --list-units` for more information.

            # Recognized sizes

            Argument of "--size" is either a couple of length (e.g. "21cm 29.7cm", "21cmx297mm", "210mm×297mm"), or a case-insensitive standardized paper format (e.g. "A4").

            For instance, A4 paper can be set using "--size=A4" or "--size=21cmx29.7cm".

            Available paper formats are: {sizes}. See `pdfautonup --list-sizes` for more information.
            """
            )
            .format(
                units=", ".join(sorted(papersize.UNITS_HELP)),
                sizes=", ".join(papersize.SIZES_HELP),
            )
            .strip()
        )
        sys.exit(0)


class HelpOrders(ArgparseAction):
    """Argparse action to display help about writing orders."""

    #: pylint: disable= line-too-long

    def __call__(self, *args, **kwargs):
        # pylint: disable=signature-differs
        print(
            textwrap.dedent(
                """\
            The order is defined as an horizontal and vertical order, for instance `lrtb` (Left to Right, Top to Bottom) for the latin (and english, french…) writing order. Available writing orders are:
            - `lrtb` (or `latin`): Left to Right, Top to Bottom,
            - `rltb` (or `arabic`): Right to Left, Top to Bottom,
            - `tbrl` (or `japanese`): Top to Bottom, Right to Left,
            - and any other combination.

            The following writing orders are also available, but are to be considered something between experimental and easter eggs (that is: they may be removed without notice if they are too much of a hassle to maintain). I do not know why you would need it, but who knows?

            # Boustrophedon

            Any order starting with `B` is a "Boustrophedon" order (see your favorite dictionary or encyclopedia for more information). Any valid combination of `B<start><direction>` is accepted, where:
            - `<start>` is the starting corner (`tl`, `tr`, `bl`, `br` for top left, top right, bottom left, bottom right),
            - `<direction>` is one of `t`, `b`, `l`, `r` (top, bottom, left, right).

            For instance, `Btlr` means: "Boustrophedon, starting on the Top Left corner, heading Right".

            # Spiral

            Any order starting with `S` is a spiral. Any combination of `S<start><rotation><direction>` is accepted, where:
            - <start> is the starting (or ending for outward spirals) corner (`tl`, `tr`, `bl`, `br` for top left, top right, bottom left, bottom right),
            - <rotation> is `c` or `a` (clockwise or anticlockwise),
            - <direction> is `i` or `o` (inward (from the border to the center) or outward (from the center to the border)).

            For instance, `Strci` is a spiral starting from the top right corner, clockwise, toward the center, and `Sblao` is a spiral ending in the bottom left corner, anticlockwise, from the center.
            """
            )
        )
        sys.exit(0)


class ListUnits(ArgparseAction):
    """Argparse action to list and describe units."""

    def __call__(self, *args, **kwargs):
        # pylint: disable=signature-differs
        for key, value in sorted(papersize.UNITS_HELP.items()):
            print(f"{key}: {value}")
        sys.exit(0)


class ListSizes(ArgparseAction):
    """Argparse action to list and describe sizes."""

    def __call__(self, *args, **kwargs):
        # pylint: disable=signature-differs
        for key, value in sorted(papersize.SIZES_HELP.items()):
            print(f"{key}: {value}")
        sys.exit(0)


def commandline_parser():
    """Return a command line parser."""
    # pylint: disable=line-too-long

    parser = argparse.ArgumentParser(
        prog="pdfautonup",
        description=textwrap.dedent(
            """\
            Convert PDF files to 'n-up' file, with multiple input pages per destination pages. The output size is configurable, and the program compute the page layout, to fit as much source pages in per destination pages as possible. If necessary, the source pages are repeated to fill all destination pages.
            """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=textwrap.dedent(
            # pylint: disable=consider-using-f-string
            """\
                The backend Python library used to read and write PDF files can be forced using the environment variable PDFBACKEND. If this variable is not defined (or defined with an invalid value), a library is automatically (and silently) chosen. Available libraries are: {pdfbackends}.
                """.format(
                pdfbackends=", ".join(
                    f"'{item}'" for item in pdfbackend.PDFBACKENDS if item != "auto"
                ),
            )
        ),
    )

    parser.add_argument(
        "--version",
        help="Show version",
        action="version",
        # pylint: disable=no-member
        version=f"%(prog)s {VERSION}\npdf backend: {pdfbackend.get_backend().VERSION}\n",
    )

    parser.add_argument(
        "--help-paper",
        help="Show an help message about paper sizes, and exit.",
        action=HelpPaper,
    )

    parser.add_argument(
        "--list-units",
        help="Display the list of available units, and their meaning, and exit.",
        action=ListUnits,
    )

    parser.add_argument(
        "--list-sizes",
        help="Display the list of available sizes, and their meaning, and exit.",
        action=ListSizes,
    )

    parser.add_argument(
        "files",
        metavar="FILES",
        help=(
            "PDF files to merge. If their page sizes are different, they are "
            "considered to have the same page size, which is the maximum width "
            """and height of all pages. To read from standard input, use "-"."""
        ),
        nargs="*",
        type=inputfile_type,
        default=[None],
    )

    parser.add_argument(
        "--output",
        "-o",
        help=(
            'Destination file (or "-" to write to standard output). Default is "-nup" appended to first source file (excepted if first source file is standard input, where default output is standard output).'
        ),
        type=str,
        nargs="?",
    )

    parser.add_argument(
        "--interactive",
        "-i",
        help="Ask before overwriting destination file if it exists.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--algorithm",
        "-a",
        help=textwrap.dedent(
            """\
            Algorithm used to arrange source documents into destination documents. This program tries to put as many copies of the source document into the destination document, given that:
            - `fuzzy`: documents can overlap, or leave blank spaces between them, but not too much;
            - `panel`: the gap length between documents is fixed, and a minimum destination margin is respected.
            """
        ),
        default=None,
        choices=["fuzzy", "panel"],
    )

    parser.add_argument(
        "--orientation",
        "-O",
        help=textwrap.dedent(
            """\
            Destination paper orientation. Default is 'auto', which choose the paper orientation to fit the maximum number of source pages on the destination page.
            """
        ),
        default="auto",
        choices=["auto", "portrait", "landscape"],
    )

    parser.add_argument(
        "--size",
        "-s",
        dest="target_size",
        help="Size of target paper. Run `pdfautonup --help-paper` for more information.",
        default=None,
        action="store",
        type=size_type,
    )

    parser.add_argument(
        "--margin",
        "-m",
        dest="margin",
        help=textwrap.dedent(
            """\
            Margin size. Either one length (e.g. "0.8cm") or up to four length, to specify different margins on each side (e.g. "1cm 2cm 1cm 3cm"). Run `pdfautonup --help-paper` for more information.
            """
        ),
        default=None,
        type=margin_type,
        action="store",
    )

    parser.add_argument(
        "--gap",
        "-g",
        help=textwrap.dedent(
            """\
            Gap size (e.g. "1.2mm"). Run `pdfautonup --help-paper` for more information about units.
            """
        ),
        default=None,
        type=length_type,
        action="store",
    )

    parser.add_argument(
        "--repeat",
        "-r",
        help=textwrap.dedent(
            """
        Number of times the input files have to be repeated. Possible values are:
        - an integer;
        - 'fit': the input files are repeated enough time to leave no blank
          space in the output file.
        - 'auto': if there is only one input page, equivalent to 'fit'; else,
          equivalent to 1.
        """
        ),
        type=repeat_type,
        default="auto",
        action="store",
    )

    parser.add_argument(
        "--progress",
        "-p",
        help=textwrap.dedent("""\
        Text to print after processing each page. Strings "{page}", "{pagetotal}", "{percent}" are replaced by their respective values. The following alias are defined:
        - 'none': no progress;
        - 'dot': '.';
        - 'pages': '{page}/{total}\\n';
        - 'percent': '{percent}%%\\n'.
        """
        ),
        type=progress_type,
        default="",
        action="store",
    )

    def _type_writing_order(string):
        if string in order.ORDERS_ALL:
            return string
        raise argparse.ArgumentTypeError(
            # pylint: disable=consider-using-f-string
            "invalid choice: '{}' (choose from {})".format(
                string, ", ".join(f"'{item}'" for item in order.ORDERS_PUBLIC)
            )
        )

    parser.add_argument(
        "--writing-order",
        "-w",
        help=textwrap.dedent(
            """\
                Writing order: define the order in which source pages should be displayed on the destination page (default `latin`). The order is defined as an horizontal and vertical order, for instance `lrtb` (Left to Right, Top to Bottom) for the latin (and english, french…) writing order. Available writing orders are:
                - `lrtb` (or `latin`): Left to Right, Top to Bottom,
                - `rltb` (or `arabic`): Right to Left, Top to Bottom,
                - `tbrl` (or `japanese`): Top to Bottom, Right to Left,
                - and any other combination.
                More exotic, useless, or experimental writing orders are available. See `--help-orders` for more information.
                """
        ),
        default="latin",
        metavar="{latin,arabic,japanese,btlr,btrl,[…],Bblr,Bblt,[…],Sblai,Sblao,[…]}",
        type=_type_writing_order,
    )

    parser.add_argument(
        "--help-orders",
        help="Show an help message about writing orders, and exit.",
        action=HelpOrders,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        help="Increase verbosity.",
        action="store_true",
    )

    return parser



# Copyright Louis Paternault 2019-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Read and write PDF files using the PyPDF library."""

import contextlib
import io

import pypdf

from . import (
    METADATA_KEYS,
    AbstractPDFFileReader,
    AbstractPDFFileWriter,
    AbstractPDFPage,
)

VERSION = "pypdf" + pypdf.__version__


def _rectangle_size(rectangle):
    """Return the dimension of rectangle (width, height)."""
    return (
        rectangle.upper_right[0] - rectangle.lower_left[0],
        rectangle.upper_right[1] - rectangle.lower_left[1],
    )


def _metadata2dict(file):
    return {
        keyword: getattr(file.metadata, keyword)
        for keyword in METADATA_KEYS
        if (getattr(file.metadata, keyword, None) is not None)
    }


class PDFFileReader(AbstractPDFFileReader):
    """Read a PDF file."""

    def __init__(self, file=None):
        super().__init__()
        self._file = pypdf.PdfReader(file)

    def close(self):
        pass

    def __len__(self):
        return len(self._file.pages)

    def __iter__(self):
        for page in self._file.pages:
            yield PDFPage(page)

    def __getitem__(self, key):
        return self._file.pages[key]

    @property
    def metadata(self):
        return _metadata2dict(self._file)


class PDFFileWriter(AbstractPDFFileWriter):
    """PDF file writer."""

    def __init__(self):
        super().__init__()
        self._file = pypdf.PdfWriter()

    @contextlib.contextmanager
    def new_page(self, width, height):
        yield PDFPage(self._file.add_blank_page(width=width, height=height))

    def write(self, file):
        if isinstance(file, io.BytesIO):
            self._file.write(file)
        else:
            with open(file, "wb") as stream:
                self._file.write(stream)

    @property
    def metadata(self):
        return _metadata2dict(self._file)

    @metadata.setter
    def metadata(self, value):
        self._file.add_metadata(
            {f"/{keyword.capitalize()}": data for keyword, data in value.items()}
        )


class PDFPage(AbstractPDFPage):
    """Page of a PDF file (using pypdf)."""

    @property
    def rotated_size(self):
        size = _rectangle_size(self._page.mediabox)
        if self._page.rotation % 180 == 0:
            return size
        return (size[1], size[0])

    def merge_translated_page(self, page, x, y):
        # Handle rotated pages
        page.transfer_rotation_to_content()

        # Merge page
        self._page.merge_transformed_page(
            page, pypdf.Transformation().translate(float(x), float(y))
        )



# Copyright Louis Paternault 2019-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract classes to read and write PDF files."""

import importlib
import os

PDFBACKENDS = ["auto", "pymupdf", "pypdf"]

METADATA_KEYS = ["title", "author", "keywords", "creator", "producer"]


def get_backend():
    """Return the pdf backend, as a module.

    The choice is done by reading the ``PDFBACKEND`` environment variable.
    """
    requested = os.environ.get("PDFBACKEND", "auto").lower()
    if requested not in PDFBACKENDS:
        requested = "auto"

    if requested == "auto":
        for name in PDFBACKENDS:
            if name == "auto":
                continue
            try:
                return importlib.import_module(f".{name}", __name__)
            except ImportError:
                continue
        raise ImportError("Could not import any pdf backend.")
    return importlib.import_module(f".{requested}", __name__)


class AbstractPDFFileReader:
    """PDF file reader."""

    def __init__(self, file=None):
        """Open file. If `file` is `None`, read from standard input."""

    def close(self):
        """Close file."""
        raise NotImplementedError()

    def __iter__(self):
        """Iterate over pages of PDF."""
        raise NotImplementedError()

    def __len__(self):
        """Return the number of pages."""
        raise NotImplementedError()

    @property
    def metadata(self):
        """Return a dictionary of PDF metadata."""
        raise NotImplementedError()

    def __getitem__(self, key):
        raise NotImplementedError()


class AbstractPDFFileWriter:
    """PDF file writer."""

    def new_page(self, width, height):
        """Create a new page, of size (width, height)."""
        raise NotImplementedError()

    @property
    def metadata(self):
        """Return file metadata (as a dictionary)."""
        raise NotImplementedError()

    @metadata.setter
    def metadata(self, value):
        raise NotImplementedError()

    def write(self, file):
        """Write file to file system."""
        raise NotImplementedError()


class AbstractPDFPage:
    """Page of a PDF file."""

    def __init__(self, page):
        super().__init__()
        self._page = page

    @property
    def rotated_size(self):
        """Return the media box size, taking into account page rotation."""
        raise NotImplementedError()

    def merge_translated_page(self, page, x, y):
        """Merge `page` into current page, at coordinates `(x, y)`."""
        raise NotImplementedError()



# Copyright Louis Paternault 2014-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Errors and exceptions"""


class PdfautonupError(Exception):
    """Generic error for pdfautonup"""

    def __init__(self, message=""):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message


class CouldNotParse(PdfautonupError):
    """Could not parse string as a paper size."""

    def __str__(self):
        return f"Could not parse paper size '{self.message}'."



# Copyright 2014-2025 Louis Paternault and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Different algorithm to fit source files into destination files."""

import contextlib
import logging
import operator
from collections import namedtuple
from decimal import Decimal, DivisionByZero

import papersize

from . import errors, order, paper, pdfbackend

LOGGER = logging.getLogger("pdfautonup")


def _dist_to_round(x):
    """Return distance of ``x`` to ``round(x)``."""
    return abs(x - round(x))


class _Layout:
    def __init__(self, target_size, **arguments):
        # pylint: disable=unused-argument
        self.target_size = target_size
        self.pdf = pdfbackend.get_backend().PDFFileWriter()
        self.order = getattr(
            order, arguments["more"].get("writingorder", order.DEFAULT_ORDER)
        )

    @contextlib.contextmanager
    def new_page(self):
        """Context manager to return a new, blank page.

        The page is expected to stay unchanged after the ``__exit__()`` method.
        """
        with self.pdf.new_page(
            width=round(self.target_size[0]), height=round(self.target_size[1])
        ) as new_page:
            yield new_page

    def add_page(self, sourcepage, destpage, number):
        """Add ``page`` to the destination file.

        It is added at the right place for `number`:
        if `N` source pages can fit into *one* destination page,
        then `number` is an integer from `0` to `N-1`.
        """
        (x, y) = self.cell_bottomleft(number)
        destpage.merge_translated_page(sourcepage, x, y)

    def set_metadata(self, metadata):
        """Set metadata on current pdf."""
        metadata["producer"] = (
            # pylint: disable=line-too-long
            f"pdfautonup, using the {pdfbackend.get_backend().__name__.split('.')[-1]} library — http://framagit.org/spalax/pdfautonup"
        )
        self.pdf.metadata = metadata

    def write(self, output):
        """Write (pseudo-)file"""
        self.pdf.write(output)

    def cell_bottomleft(self, num):
        """Return the bottom left coordinate of ``num``th cell of page."""
        raise NotImplementedError()

    @property
    def pages_per_page(self):
        """Return the number of source pages per destination page."""
        raise NotImplementedError()


class Fuzzy(_Layout):
    """Documents can overlap, and space can be wasted, but not too much."""

    #: A target size, associated with the number of source pages that will fit
    #: in it, per width and height (``cell_number[0]`` and ``cell_number[1]``).
    Grid = namedtuple("Grid", ["cell_number", "target_size", "margins", "gaps"])

    def __init__(self, source_size, target_size, **arguments):
        if (
            arguments["more"].get("margin", None) is not None
            or arguments["more"].get("gap", None) is not None
        ):
            LOGGER.warning(
                "Arguments `--margin` and `--gap` are ignored with algorithm `fuzzy`."
            )
        self.source_size = source_size
        if arguments["orientation"] == "landscape":
            self.grid = self.generate_grid(
                source_size, papersize.rotate(target_size, papersize.LANDSCAPE)
            )
        elif arguments["orientation"] == "portrait":
            self.grid = self.generate_grid(
                source_size, papersize.rotate(target_size, papersize.PORTRAIT)
            )
        else:
            self.grid = min(
                self.generate_grid(source_size, target_size),
                self.generate_grid(source_size, (target_size[1], target_size[0])),
                key=self.ugliness,
            )

        super().__init__(self.grid.target_size, **arguments)

    def ugliness(self, grid):
        """Return the "ugliness" of this ``grid``.

        - A layout that fits perfectly has an ugliness of 0.
        - The maximum ugliness is 1.
        """
        target_width, target_height = grid.target_size
        source_width, source_height = self.source_size
        return (
            _dist_to_round(target_width / source_width) ** 2
            + _dist_to_round(target_height / source_height) ** 2
        )

    @staticmethod
    def _margins(target_size, source_size, cell_number):
        if cell_number[0] == 1:
            width = (target_size[0] - source_size[0] * cell_number[0]) / 2
        else:
            width = 0
        if cell_number[1] == 1:
            height = (target_size[1] - source_size[1] * cell_number[1]) / 2
        else:
            height = 0
        return [width, height]

    @staticmethod
    def _gaps(target_size, source_size, cell_number):
        if cell_number[0] == 1:
            width = Decimal(0)
        else:
            width = (target_size[0] - cell_number[0] * source_size[0]) / (
                cell_number[0] - 1
            )
        if cell_number[1] == 1:
            height = Decimal(0)
        else:
            height = (target_size[1] - cell_number[1] * source_size[1]) / (
                cell_number[1] - 1
            )
        return (width, height)

    def generate_grid(self, source_size, target_size):
        """Return a :class:`self.Grid` object for arguments.

        The main function is computing the number of source pages per
        destination pages.
        """
        try:
            cell_number = (
                max(1, round(target_size[0] / source_size[0])),
                max(1, round(target_size[1] / source_size[1])),
            )
        except DivisionByZero as error:
            raise errors.PdfautonupError(
                "Error: A PDF page have a null dimension."
            ) from error
        return self.Grid(
            cell_number,
            target_size,
            self._margins(target_size, source_size, cell_number),
            self._gaps(target_size, source_size, cell_number),
        )

    def cell_bottomleft(self, num):
        row, column = self.order(num, *self.grid.cell_number)
        return (
            self.grid.margins[0] + (self.source_size[0] + self.grid.gaps[0]) * row,
            self.grid.margins[1] + (self.source_size[1] + self.grid.gaps[1]) * column,
        )

    @property
    def pages_per_page(self):
        return self.grid.cell_number[0] * self.grid.cell_number[1]


class Panelize(_Layout):
    """Minimum margin is defined, as well as fixed gap."""

    #: Define how the source page will fit into the destination page.
    #: - `margin` is the destination margin of the bottom left corner
    #:   (including wasted space), as a tuple: (left, bottom).
    #: - `sourcex` is the 'extended' source size (source size, together with gap).
    Grid = namedtuple(
        "Grid", ["margin", "sourcex", "dimension", "target", "pagenumber"]
    )

    def __init__(self, source_size, target_size, **arguments):
        # pylint: disable=too-many-arguments
        self.gap = arguments["more"].get("gap", papersize.parse_length("0"))
        if self.gap is None:
            self.gap = papersize.parse_length("0")
        elif isinstance(self.gap, str):
            self.gap = papersize.parse_length(self.gap)

        self.margin = paper.Margins.parse(arguments["more"].get("margin", None))

        if arguments["orientation"] == "landscape":
            self.grid = self.generate_grid(
                source_size,
                papersize.rotate(target_size, papersize.LANDSCAPE),
                self.margin,
            )
        elif arguments["orientation"] == "portrait":
            self.grid = self.generate_grid(
                source_size,
                papersize.rotate(target_size, papersize.PORTRAIT),
                self.margin,
            )
        else:
            self.grid = max(
                self.generate_grid(source_size, target_size, self.margin),
                self.generate_grid(
                    source_size, (target_size[1], target_size[0]), self.margin
                ),
                key=operator.attrgetter("pagenumber"),
            )

        if self.pages_per_page == 0:
            raise errors.PdfautonupError(
                "Error: Format constraints too tight: Cannot fit any"
                "source page into destination page."
            )

        super().__init__(self.grid.target, **arguments)

    def generate_grid(
        self,
        source: tuple[Decimal, Decimal],
        target: tuple[Decimal, Decimal],
        margin: paper.Margins,
    ) -> "Panelize.Grid":
        """Computes the geometry of the output file.

        It computes how many source pages will be merged on each destination pages,
        the total number of pages, the margins, etc.
        """

        dimension = (
            self._num_fit(target[0], margin.left + margin.right, source[0]),
            self._num_fit(target[1], margin.top + margin.bottom, source[1]),
        )

        wasted = (
            self._wasted(
                target[0], margin.left + margin.right, dimension[0], source[0]
            ),
            self._wasted(
                target[1], margin.top + margin.bottom, dimension[1], source[1]
            ),
        )
        return self.Grid(
            margin=(margin.left + wasted[0], margin.bottom + wasted[1]),
            sourcex=(source[0] + self.gap, source[1] + self.gap),
            dimension=dimension,
            target=target,
            pagenumber=dimension[0] * dimension[1],
        )

    def _wasted(self, dest, margins, num, source):
        """Return the amount of wasted space

        When fitting `num` elements of size `source` into element of size
        `dest` (in one dimension).

        `margins` is the total margin in the beginning and the end of this dimension.
        """
        return (dest - num * (source + self.gap) - margins + self.gap) / 2

    def _num_fit(self, target, margins, source):
        """Return the number of source elements that can fit in target.

        `source`, `target` and `margins` are all sizes, in one dimension.

        `margins` is the total margin in the beginning and the end of this dimension.
        """
        return int((target - margins + self.gap) // (source + self.gap))

    @property
    def pages_per_page(self):
        return self.grid.dimension[0] * self.grid.dimension[1]

    def cell_bottomleft(self, num):
        row, column = self.order(num, *self.grid.dimension)
        return (
            self.grid.margin[0] + self.grid.sourcex[0] * row,
            self.grid.margin[1] + self.grid.sourcex[1] * column,
        )



#!/usr/bin/env python3

# Copyright Louis Paternault 2014-2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Main function for the command."""

import io
import logging
import os
import sys

from . import errors, options, pdfautonup, pdfbackend

LOGGER = logging.getLogger("pdfautonup")


def _progress_printer(string):
    """Returns a function that prints the progress message."""

    def print_progress(page, total):
        """Print progress message."""
        try:
            text = string.format(
                page=page, total=total, percent=int(page * 100 / total)
            )
        except:  # pylint: disable=bare-except
            text = string
        print(text, end="")
        sys.stdout.flush()

    return print_progress


def _process_output(file, output):
    if output == "-" or (output is None and not isinstance(file, str)):
        return io.BytesIO()
    if isinstance(output, str):
        return output
    if output is None:
        if file.endswith(".pdf"):
            return f"""{file[: -len(".pdf")]}-nup.pdf"""
        return f"{file}-nup.pdf"
    raise TypeError()


def main():
    """Main function"""
    try:
        arguments = options.commandline_parser().parse_args(sys.argv[1:])

        if arguments.verbose:
            # pylint: disable=no-member
            sys.stderr.write(f"Using pdf backend {pdfbackend.get_backend().VERSION}\n")

        # Process input file names
        if None in arguments.files and arguments.interactive:
            LOGGER.error(
                """Cannot ask user input while reading files from standard input. """
                """Try removing the "--interactive" (or "-i") option."""
            )
            sys.exit(1)
        for i, file in enumerate(arguments.files):
            if file is None:
                arguments.files[i] = io.BytesIO(sys.stdin.buffer.read())

        # Compute output file name
        output = _process_output(arguments.files[0], arguments.output)
        if isinstance(output, str) and arguments.interactive and os.path.exists(output):
            if (
                input(f"File {output} already exists. Overwrite (y/[n])? ").lower()
                != "y"
            ):
                raise errors.PdfautonupError("Cancelled by user.")

        pdfautonup(
            arguments.files,
            output,
            arguments.target_size,
            algorithm=arguments.algorithm,
            repeat=arguments.repeat,
            orientation=arguments.orientation,
            progress=_progress_printer(arguments.progress),
            more={
                "gap": arguments.gap,
                "margin": arguments.margin,
                "writingorder": arguments.writing_order,
            },
        )
        if not (arguments.progress.endswith("\n") or arguments.progress == ""):
            print()
        if isinstance(output, io.BytesIO):
            sys.stdout.buffer.write(output.getvalue())
    except KeyboardInterrupt:
        print()
        sys.exit(1)
    except errors.PdfautonupError as error:
        LOGGER.error(error)
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()



# Copyright Louis Paternault 2025
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# Yogroup=u should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Writing orders"""

import functools
import itertools
import random as _random

DEFAULT_ORDER = "latin"

# There is probably a smart way to format those function in a single one, but,
# this one is clear, and I am lazy…

################################################################################
# Officially supported writing orders


def lrtb(n, width, height):
    """Left to right, top to bottom (english, french, etc. writing order)."""
    return n % width, (width * height - n - 1) // width


def lrbt(n, width, height):
    """Left to right, bottom to top"""
    # pylint: disable=unused-argument
    return n % width, n // width


def rltb(n, width, height):
    """Right to left, top to bottom (arabic writing order)."""
    return (width * height - n - 1) % width, (width * height - n - 1) // width


def rlbt(n, width, height):
    """Right to left, bottom to top."""
    return (width * height - n - 1) % width, n // width


def tblr(n, width, height):
    """Top to bottom, left to right."""
    return n // height, (width * height - n - 1) % height


def btlr(n, width, height):
    """Bottom to top, left to right"""
    # pylint: disable=unused-argument
    return n // height, n % height


def tbrl(n, width, height):
    """Top to bottom, right to left (japanese writing order)."""
    return (width * height - n - 1) // height, (width * height - n - 1) % height


def btrl(n, width, height):
    """Bottom to top, right to left ."""
    return (width * height - n - 1) // height, n % height


################################################################################
# Some aliases


def latin(*args, **kwargs):
    """Latin writing order (alias for `lrtb`)."""
    return lrtb(*args, **kwargs)


def japanese(*args, **kwargs):
    """Japanese writing order (alias for `tbrl`)."""
    return tbrl(*args, **kwargs)


def arabic(*args, **kwargs):
    """Arabic writing order (alias for `rltb`)."""
    return rltb(*args, **kwargs)


################################################################################
# Boustrophedon (experimental)
# This is useless, but not as useless as the spirals below!

# pylint: disable=invalid-name, no-else-return


def Btrl(n, width, height):
    """Boustrophedon, starting from the top right corner, and heading left."""
    if n // width % 2 == 0:
        return rltb(n, width, height)
    else:
        return lrtb(n, width, height)


def Btlr(n, width, height):
    """Boustrophedon, starting from the top left corner, and heading right."""
    if n // width % 2 == 0:
        return lrtb(n, width, height)
    else:
        return rltb(n, width, height)


def Bblr(n, width, height):
    """Boustrophedon, starting from the bottom left corner, and heading right."""
    if n // width % 2 == 0:
        return lrbt(n, width, height)
    else:
        return rlbt(n, width, height)


def Bblt(n, width, height):
    """Boustrophedon, starting from the bottom left corner, and heading top."""
    if n // height % 2 == 0:
        return btlr(n, width, height)
    else:
        return tblr(n, width, height)


def Bbrl(n, width, height):
    """Boustrophedon, starting from the bottom right corner, and heading left."""
    if n // width % 2 == 0:
        return rlbt(n, width, height)
    else:
        return lrbt(n, width, height)


def Bbrt(n, width, height):
    """Boustrophedon, starting from the bottom right corner, and heading top."""
    if n // height % 2 == 0:
        return btrl(n, width, height)
    else:
        return tbrl(n, width, height)


def Btlb(n, width, height):
    """Boustrophedon, starting from the top left corner, and heading bottom."""
    if n // height % 2 == 0:
        return tblr(n, width, height)
    else:
        return btlr(n, width, height)


def Btrb(n, width, height):
    """Boustrophedon, starting from the top right corner, and heading bottom."""
    if n // height % 2 == 0:
        return tbrl(n, width, height)
    else:
        return btrl(n, width, height)


################################################################################
# Spiral (experimental)
# Why did I even waste time coding this?


def _rotation(  # pylint: disable=too-many-return-statements, inconsistent-return-statements
    start: tuple[int, int], clockwise: bool
) -> tuple[tuple[int, int]]:
    """Return the list of four directions, that are to be cycled to rotate.

    For instance: right, bottom, left, top as coordinates of the direction:
    `((1, 0), (0, -1), (-1, 0), (0, 1))`

    :param start: Coordinates of the starting corner.
        Any non-zero value is interpreted as "as right/top as possible".
    :param clockwise: If `True`, rotate clockwise. Otherwise, rotate anticlockwise.
    """
    # If I was smarter, or less tired, or less lazy, I would have found a smarter way to do this.
    # But this is legible and clear, so who cares?
    if clockwise and start[0] == 0 and start[1] == 0:
        # Bottom left corner
        return ((0, 1), (1, 0), (0, -1), (-1, 0))
    if clockwise and start[0] != 0 and start[1] == 0:
        # Bottom right corner
        return ((-1, 0), (0, 1), (1, 0), (0, -1))
    if clockwise and start[0] == 0 and start[1] != 0:
        # Top left corner
        return ((1, 0), (0, -1), (-1, 0), (0, 1))
    if clockwise and start[0] != 0 and start[1] != 0:
        # Top right corner
        return ((0, -1), (-1, 0), (0, 1), (1, 0))
    if not clockwise and start[0] == 0 and start[1] == 0:
        # Bottom left corner
        return ((1, 0), (0, 1), (-1, 0), (0, -1))
    if not clockwise and start[0] != 0 and start[1] == 0:
        # Bottom right corner
        return ((0, 1), (-1, 0), (0, -1), (1, 0))
    if not clockwise and start[0] == 0 and start[1] != 0:
        # Top left corner
        return ((0, -1), (1, 0), (0, 1), (-1, 0))
    if not clockwise and start[0] != 0 and start[1] != 0:
        # Top right corner
        return ((-1, 0), (0, -1), (1, 0), (0, 1))


@functools.lru_cache(1)
def _spiral(
    size: tuple[int, int], start: tuple[int, int], clockwise: bool, inward: bool
) -> dict[int, tuple[int, int]]:
    """Return the mapping from integer to coordinates, in a spiral.

    Suppose `size == (4, 3)`. This function builds the following spiral:

    +--+--+--+--+
    | 0| 1| 2| 3|
    +--+--+--+--+
    | 9|10|11| 4|
    +--+--+--+--+
    | 8| 7| 6| 5|
    +--+--+--+--+

    And then returns the dictionary mapping integers (those you can see in the spiral)
    to coordinates (the coordinate of the number in the spiral).
    For instance, if the return value is stored in `spiral`, then:
    `spiral[10] == (2, 1)` and `spiral[1] == (1, 2)`.

    :param size: Size of the grid, as `(width, height)`.
    :param start: Initial (or last, with `inward=False`) position of the spiral, as coordinates.
        If those coordinates are not a corner of the grid, the result is undefined.
    :param clockwise: If `True`, rotate clockwise. Otherwise, rotate anticlockwise.
    :param inward: If `True`, the spiral starts from a corner, and spirals to the conter.
        If `False`, the spirals goes the opposite.
        In this latter case, the `start` argument is the *end* of the spiral, not the start.
    """
    # Build grid
    if not inward:
        clockwise = not clockwise
    grid = [[None] * size[1] for _ in range(size[0])]
    rotation = itertools.cycle(_rotation(start, clockwise))
    direction = next(rotation)
    for i in range(size[0] * size[1]):
        grid[start[0]][start[1]] = i
        try:
            if grid[start[0] + direction[0]][start[1] + direction[1]] is not None:
                direction = next(rotation)
        except IndexError:
            direction = next(rotation)
        start = (
            start[0] + direction[0],
            start[1] + direction[1],
        )
        # Debug: uncomment to display the grid
        # print("\n".join(str(line) for line in zip(*grid)))

    # Grid is built. Return a reverse-map of it:
    # given a number, we get its position in the grid.
    # pylint: disable=no-else-return
    if inward:
        return {
            grid[x][y]: (x, y) for x in range(len(grid)) for y in range(len(grid[x]))
        }
    else:
        # We "reverse" the grid numbers
        return {
            size[0] * size[1] - grid[x][y] - 1: (x, y)
            for x in range(len(grid))
            for y in range(len(grid[x]))
        }


def Stlci(n, width, height):
    """Spiral, starting from the top left corner, clockwise, inward."""
    return _spiral((width, height), (0, height - 1), clockwise=True, inward=True)[n]


def Stlco(n, width, height):
    """Spiral, starting from the top left corner, clockwise, outward."""
    return _spiral((width, height), (0, height - 1), clockwise=False, inward=False)[n]


def Sblai(n, width, height):
    """Spiral, starting from the bottom left corner, anticlockwise, inward."""
    return _spiral((width, height), (0, 0), clockwise=False, inward=True)[n]


def Sblao(n, width, height):
    """Spiral, starting from the bottom left corner, anticlockwise, outward."""
    return _spiral((width, height), (0, 0), clockwise=False, inward=False)[n]


def Sblci(n, width, height):
    """Spiral, starting from the bottom left corner, clockwise, inward."""
    return _spiral((width, height), (0, 0), clockwise=True, inward=True)[n]


def Sblco(n, width, height):
    """Spiral, starting from the bottom left corner, clockwise, outward."""
    return _spiral((width, height), (0, 0), clockwise=True, inward=False)[n]


def Sbrai(n, width, height):
    """Spiral, starting from the bottom right corner, anticlockwise, inward."""
    return _spiral((width, height), (width - 1, 0), clockwise=False, inward=True)[n]


def Sbrao(n, width, height):
    """Spiral, starting from the bottom right corner, anticlockwise, outward."""
    return _spiral((width, height), (width - 1, 0), clockwise=False, inward=False)[n]


def Sbrci(n, width, height):
    """Spiral, starting from the bottom right corner, clockwise, inward."""
    return _spiral((width, height), (width - 1, 0), clockwise=True, inward=True)[n]


def Sbrco(n, width, height):
    """Spiral, starting from the bottom right corner, clockwise, outward."""
    return _spiral((width, height), (width - 1, 0), clockwise=True, inward=False)[n]


def Stlai(n, width, height):
    """Spiral, starting from the top left corner, anticlockwise, inward."""
    return _spiral((width, height), (0, height - 1), clockwise=False, inward=True)[n]


def Stlao(n, width, height):
    """Spiral, starting from the top left corner, anticlockwise, outward."""
    return _spiral((width, height), (0, height - 1), clockwise=False, inward=False)[n]


def Strai(n, width, height):
    """Spiral, starting from the top right corner, anticlockwise, inward."""
    return _spiral(
        (width, height), (width - 1, height - 1), clockwise=False, inward=True
    )[n]


def Strao(n, width, height):
    """Spiral, starting from the top right corner, anticlockwise, outward."""
    return _spiral(
        (width, height), (width - 1, height - 1), clockwise=False, inward=False
    )[n]


def Strci(n, width, height):
    """Spiral, starting from the top right corner, clockwise, inward."""
    return _spiral(
        (width, height), (width - 1, height - 1), clockwise=True, inward=True
    )[n]


def Strco(n, width, height):
    """Spiral, starting from the top right corner, clockwise, outward."""
    return _spiral(
        (width, height), (width - 1, height - 1), clockwise=True, inward=False
    )[n]


# Random (easter egg)

SEED = _random.random()


@functools.lru_cache(1)
def _shuffle(size):
    """Return a randomly shuffled list of integers from 0 to `size-1`.

    This function is cached and deterministic.
    """
    _random.seed(f"{SEED} {size}")
    return _random.sample(range(size), k=size)


def random(n, width, height):
    """Random writing order"""
    # This cannot be totally random, otherwise some pages will overlap.
    return latin(_shuffle(width * height)[n], width, height)


ORDERS = [
    function.__name__
    for function in (
        # Officially supported
        btlr,
        btrl,
        lrbt,
        lrtb,
        rlbt,
        rltb,
        tblr,
        tbrl,
    )
]

ORDERS_ALIASES = [
    function.__name__
    for function in (
        # Aliases
        arabic,
        japanese,
        latin,
    )
]

ORDERS_EXPERIMENTAL = [
    function.__name__
    for function in (
        ## Boustrophédon (experimental)
        Bblr,
        Bblt,
        Bbrl,
        Bbrt,
        Btlb,
        Btlr,
        Btrb,
        Btrl,
        ## Spiral (experimental)
        Sblai,
        Sblao,
        Sblci,
        Sblco,
        Sbrai,
        Sbrao,
        Sbrci,
        Sbrco,
        Stlai,
        Stlao,
        Stlci,
        Stlco,
        Strai,
        Strao,
        Strci,
        Strco,
    )
]

ORDERS_EASTEREGGS = [function.__name__ for function in (random,)]

ORDERS_PUBLIC = ORDERS_ALIASES + ORDERS + ORDERS_EXPERIMENTAL
ORDERS_ALL = ORDERS_PUBLIC + ORDERS_EASTEREGGS
